package filegateway

import (
	"fmt"
	"io"
	"path"

	stdlib "orus.io/xbus/xbus-stdlib/lib"
	nrpc "github.com/nats-rpc/gogo-nrpc"
	xbus "xbus.io/go-xbus"
)

/*
```yaml
actor:
  kind: consumer
  service:
    type: file-gateway
    settings:
       uri: scp://remote/directory/
       secrets: path/to/passwords-file.yaml
       poll-delay: 10s
       wait-for: deletion
```
*/

// Consumer save the received files to a filesystem
type Consumer struct {
	fs        Filesystem
	targetDir string

	temporaryErrors bool
}

// NewConsumer creates a Consumer
func NewConsumer(actor *xbus.Actor) xbus.ActorService {
	options, err := LoadFilesystemOptions(actor, "target")
	if err != nil {
		actor.Log.Errorf("Could not load 'target' setting: %s", err)
		return nil
	}
	fs, err := GetFilesystem(options.Scheme, options)
	if err != nil {
		if actor.Log != nil {
			actor.Log.Errorf("Could not get filesystem: %s", err)
		} else {
			fmt.Println("Could not get filesystem:", err)
		}

		return nil
	}
	return &Consumer{
		fs:              fs,
		targetDir:       options.Path,
		temporaryErrors: actor.GetBoolSettingD("temporary-errors", false),
	}
}

func (c *Consumer) fsError(op string, err error) error {
	if c.temporaryErrors {
		return &nrpc.Error{
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: fmt.Sprintf("[filegateway] Error %s file: %s.", op, err),
		}
	}
	return err
}

// Process ...
func (c *Consumer) Process(apc *xbus.ActorProcessingContext) error {
	reader, err := apc.ReadEnvelope("default")
	if err != nil {
		return err
	}
	eventTypes := reader.EventTypes()
	if len(eventTypes) != 1 {
		return fmt.Errorf("Can only handle envelopes with 1 message, not %d", len(eventTypes))
	}
	fReader := stdlib.NewFileMessageReader(reader.GetItemChan(eventTypes[0]))
	meta, err := fReader.Metadata()
	if err != nil {
		return err
	}
	body, err := fReader.Body()
	if err != nil {
		return err
	}
	targetPath := path.Join(c.targetDir, meta.Filename)
	writer, err := c.fs.Write(targetPath, -1)
	if err != nil {
		return c.fsError("opening", err)
	}
	if _, err = io.Copy(writer, body); err != nil {
		return c.fsError("writing", err)
	}
	if err := writer.Close(); err != nil {
		return c.fsError("closing", err)
	}
	return nil
}

func init() {
	xbus.RegisterActorService("file-gateway-consumer", NewConsumer)
}
