package filegateway

import (
	"bytes"
	"context"
	"encoding/json"
	"fmt"
	"io"
	"log"
	"path"
	"path/filepath"
	"strings"
	"sync"
	"text/template"

	stdlib "orus.io/xbus/xbus-stdlib/lib"
	xbus "xbus.io/go-xbus"
	"xbus.io/go-xbus/api"
)

// RenameTargetVars holds the variables available when rendering a rename target
// path template
type RenameTargetVars struct {
	Full       string
	Dir        string
	Base       string
	Ext        string
	WithoutExt string

	EnvelopeID string
}

type openOutputFunc func(
	context *api.ProcessingContext, output string, eventTypes ...string,
) *xbus.EnvelopeSender

// Emitter watches a pattern, emit the matching files into the bus, and delete
// the file after emission
type Emitter struct {
	openOutput     openOutputFunc
	log            xbus.Logger
	fs             Filesystem
	pattern        string
	msgType        string
	postAction     string
	renameTemplate *template.Template

	exit chan struct{}
	wg   sync.WaitGroup
}

// NewEmitter creates a Emitter
func NewEmitter(actor *xbus.Actor) xbus.ActorService {
	options, err := LoadFilesystemOptions(actor, "source")
	if err != nil {
		actor.Log.Errorf("Could not load 'source': %s", err)
		return nil
	}
	fs, err := GetFilesystem(options.Scheme, options)
	if err != nil {
		actor.Log.Errorf("Could not get 'source' filesystem: %s", err)
		return nil
	}
	emitter := Emitter{
		openOutput: actor.OpenOutput,
		log:        actor.Client.Log,
		fs:         fs,
		pattern:    options.Path,
		msgType:    actor.MustGetStringSetting("message-type"),
	}
	switch postAction := actor.GetStringSetting("post-action"); {
	case postAction == "":
		emitter.postAction = "delete"
	case postAction == "delete":
		emitter.postAction = "delete"
	case strings.HasPrefix(postAction, "rename "):
		emitter.postAction = "rename"
		t, err := template.New("rename target").Parse(postAction[7:])
		if err != nil {
			actor.Log.Errorf("post-action: error parsing rename target: %s", err)
			return err
		}
		emitter.renameTemplate = t
	default:
		actor.Log.Errorf("Invalid 'post-action': %s", postAction)
		return nil
	}
	log.Println("post-action:", emitter.postAction)
	return &emitter
}

// Startup lauch the observe routine
func (e *Emitter) Startup() error {
	e.exit = make(chan struct{})
	e.wg.Add(1)

	ctx, cancel := context.WithCancel(context.Background())

	c, err := e.fs.Observe(ctx, e.pattern, e.log)
	if err != nil {
		cancel()
		return err
	}

	go e.run(ctx, cancel, c)

	return nil
}

// Shutdown stops the observe routine
func (e *Emitter) Shutdown() error {
	close(e.exit)
	e.wg.Wait()
	return nil
}

func (e *Emitter) run(ctx context.Context, ctxCancel func(), c <-chan string) {
	defer e.wg.Done()
	defer ctxCancel()

	for {
		select {
		case fname := <-c:
			if err := e.emit(fname); err != nil {
				e.log.Errorf("Could not emit %s, got: %s", fname, err)
			}
		case <-e.exit:
			return
		case <-ctx.Done():
			return
		}
	}
}

func (e *Emitter) doPostAction(fpath string, envelopeID api.UUID) error {
	switch e.postAction {
	case "delete":
		return e.fs.Delete(fpath)
	case "rename":
		buf := bytes.NewBuffer(nil)
		base := filepath.Base(fpath)
		if err := e.renameTemplate.Execute(buf, RenameTargetVars{
			Full:       fpath,
			Dir:        filepath.Dir(fpath),
			Base:       base,
			Ext:        filepath.Ext(fpath),
			WithoutExt: base[:len(base)-len(filepath.Ext(base))],
			EnvelopeID: envelopeID.String(),
		}); err != nil {
			return err
		}
		return e.fs.Rename(fpath, buf.String())
	}
	return fmt.Errorf("invalid post-action: %s", e.postAction)
}

func (e *Emitter) emit(fpath string) error {
	reader, err := e.fs.Read(fpath)
	if err != nil {
		return err
	}
	meta := stdlib.FileMeta{
		Filename: path.Base(fpath),
	}
	b, err := json.Marshal(meta)
	if err != nil {
		return err
	}
	out := e.openOutput(nil, "default", e.msgType)
	envelopeID := out.GetID()
	if err := out.AddItems(e.msgType, b); err != nil {
		return nil
	}
	for {
		var b = make([]byte, 200*1024)
		n, err := reader.Read(b)
		if n != 0 {
			if err := out.AddItems(e.msgType, b[:n]); err != nil {
				return nil
			}
		}
		if err == io.EOF {
			break
		}
		if err != nil {
			// TODO see how we can mark the emission as failed
			return err
		}
	}
	if err := out.Close(); err != nil {
		return err
	}
	return e.doPostAction(fpath, envelopeID)
}

func init() {
	xbus.RegisterActorService("file-gateway-emitter", NewEmitter)
}
