package filegateway

import (
	"context"
	"crypto/tls"
	"crypto/x509"
	"encoding/pem"
	"fmt"
	"io"
	"log"
	"os"
	"path"
	"strings"
	"sync"
	"time"

	xbus "xbus.io/go-xbus"
	"github.com/secsy/goftp"
)

// FTPFlavor is the type of the remote FTP server
type FTPFlavor int

const (
	// FTPFlavorDefault is for modern ftp servers that implements the MLSD command
	FTPFlavorDefault FTPFlavor = iota
	// FTPFlavorVSFTP is for a vsftp server
	FTPFlavorVSFTP
)

func loadServerCert(cert string) (*x509.Certificate, error) {
	pemBlock := []byte(cert)

	var derBlock *pem.Block
	for {
		derBlock, pemBlock = pem.Decode(pemBlock)
		if derBlock == nil {
			break
		}
		if derBlock.Type == "CERTIFICATE" {
			break
		}
		derBlock = nil
	}
	if derBlock == nil {
		return nil, fmt.Errorf("Found no certificate")
	}
	return x509.ParseCertificate(derBlock.Bytes)
}

func ftpFilesystemFactory(options FilesystemOptions) (Filesystem, error) {
	if options.Host == "" {
		return nil, fmt.Errorf(
			"Invalid host for a ftp(s):// uri. Expects a non-mpty host")
	}
	if options.Port == "" {
		options.Port = "21"
	}
	fs := FtpFilesystem{
		FilesystemOptions: options,
		address:           options.Host + ":" + options.Port,
		clientConfig: goftp.Config{
			User:     options.Username,
			Password: options.Password,
		},
	}
	if err := options.CustomStringD("basedir", &(fs.basedir), "/"); err != nil {
		return nil, err
	}
	if err := options.CustomIntD("maxconn", &fs.clientConfig.ConnectionsPerHost, 4); err != nil {
		return nil, err
	}

	if value, err := options.GetCustomStringD("flavor", "default"); err == nil {
		switch value {
		case "default":
			fs.flavor = FTPFlavorDefault
		case "vsftp":
			fallthrough
		case "vsftpd":
			fs.flavor = FTPFlavorVSFTP
		}
	} else {
		return nil, err
	}

	if options.Scheme == "ftps" {
		fs.clientConfig.TLSConfig = &tls.Config{
			ServerName: options.Host,
		}

		if tlsmode, err := options.GetCustomStringD("tls-mode", "explicit"); err == nil {
			switch tlsmode {
			case "explicit":
				fs.clientConfig.TLSMode = goftp.TLSExplicit
			case "implicit":
				fs.clientConfig.TLSMode = goftp.TLSImplicit
			default:
				return nil, fmt.Errorf("Invalid tls-mode. Expects \"explicit\" or \"implicit\", got %s", tlsmode)
			}
		} else {
			return nil, err
		}

		if servername, err := options.GetCustomStringD("tls-servername", ""); err == nil {
			if servername != "" {
				fs.clientConfig.TLSConfig.ServerName = servername
			}
		} else {
			return nil, err
		}

		if hostcert, err := options.GetCustomStringD("host-cert", ""); err == nil {
			if hostcert == "" {
				return nil, fmt.Errorf("'host-cert' is empty")
			}
			cert, err := loadServerCert(hostcert)
			if err != nil {
				return nil, fmt.Errorf("Error parsing 'host-cert': %s", err)
			}
			certpool := x509.NewCertPool()
			certpool.AddCert(cert)

			fs.clientConfig.TLSConfig.RootCAs = certpool
		} else {
			return nil, err
		}
	}
	client, err := goftp.DialConfig(fs.clientConfig, fs.address)
	if err != nil {
		return nil, err
	}
	fs.client = client
	return &fs, nil
}

// FtpFilesystem stores files on ftp(s) server
type FtpFilesystem struct {
	FilesystemOptions
	address      string
	clientConfig goftp.Config
	basedir      string
	flavor       FTPFlavor

	client *goftp.Client
}

func ftpErrIsNotExist(err error) bool {
	if ge, ok := err.(goftp.Error); ok && ge.Code() == 550 {
		return true
	}
	return false
}

func ftpConvertError(err error) error {
	if ftpErrIsNotExist(err) {
		return os.ErrNotExist
	}
	return err
}

func (fs *FtpFilesystem) mkdirAll(dirpath string) error {
	switch fs.flavor {
	case FTPFlavorVSFTP:
		return fs.mkdirAllNoStat(dirpath)
	default:
		return fs.mkdirAllStat(dirpath)
	}
}

func (fs *FtpFilesystem) mkdirAllStat(dirpath string) error {
	dirinfo, err := fs.client.Stat(dirpath)
	if err != nil {
		if ftpErrIsNotExist(err) {
			if err := fs.mkdirAllStat(path.Dir(dirpath)); err != nil {
				return err
			}
			if _, err := fs.client.Mkdir(dirpath); err != nil {
				return err
			}
		}
	} else if !dirinfo.IsDir() {
		return fmt.Errorf("%s is not a dir", dirpath)
	}
	return nil
}

func (fs *FtpFilesystem) mkdirAllNoStat(dirpath string) error {
	parent, name := path.Split(dirpath)
	parent = strings.TrimRight(parent, "/")

	if parent != "" && parent != "." && parent != "~" {
		fs.mkdirAllNoStat(parent)
	}

	if parent == "" {
		parent = "/"
	}

	// Because we cannot use Stat(), we need to list the parent directory
	list, err := fs.client.ReadDir(parent)
	if err != nil {
		return err
	}

	for _, info := range list {
		if info.Name() == name {
			if info.IsDir() {
				return nil
			}
			return fmt.Errorf("%s is not a directory", dirpath)
		}
	}

	if _, err := fs.client.Mkdir(dirpath); err != nil {
		return err
	}

	// Check that the directory was properly created
	list, err = fs.client.ReadDir(parent)
	if err != nil {
		return err
	}

	for _, info := range list {
		if info.Name() == name {
			if info.IsDir() {
				return nil
			}
			return fmt.Errorf("%s is not a directory", dirpath)
		}
	}

	return fmt.Errorf("Mkdir failed (no more detail available): %s", dirpath)
}

// Write ...
func (fs *FtpFilesystem) Write(filepath string, size int) (io.WriteCloser, error) {
	filepath = path.Join(fs.basedir, filepath)

	reader, writer := io.Pipe()

	filepath = path.Join(fs.basedir, filepath)

	if err := fs.mkdirAll(path.Dir(filepath)); err != nil {
		return nil, err
	}
	go func() {
		err := fs.client.Store(filepath, reader)
		reader.CloseWithError(err)
	}()
	return writer, nil
}

type ftpFileReader struct {
	nextRead  []byte
	queue     chan []byte
	Started   chan struct{}
	closed    chan struct{}
	err       error
	m         sync.Mutex
	isStarted bool
}

func newFtpFileReader() *ftpFileReader {
	return &ftpFileReader{
		queue:   make(chan []byte, 1),
		closed:  make(chan struct{}),
		Started: make(chan struct{}),
	}
}

func (fr *ftpFileReader) Err() (err error) {
	fr.m.Lock()
	err = fr.err
	fr.m.Unlock()
	return
}

func (fr *ftpFileReader) Write(buf []byte) (int, error) {
	// write to the queue
	select {
	case <-fr.closed:
		return 0, io.ErrClosedPipe
	case fr.queue <- buf:
		if !fr.isStarted {
			fr.isStarted = true
			close(fr.Started)
		}
		return len(buf), nil
	}
}

func (fr *ftpFileReader) Read(buf []byte) (int, error) {
	if len(fr.nextRead) == 0 {
		select {
		case <-fr.closed:
			return 0, fr.err
		case fr.nextRead = <-fr.queue:
		}
	}
	var chunk []byte
	if len(fr.nextRead) <= len(buf) {
		chunk = fr.nextRead
		fr.nextRead = nil
	} else {
		chunk = fr.nextRead[:len(buf)]
		fr.nextRead = fr.nextRead[len(buf):]
	}
	return copy(buf, chunk), nil
}

func (fr *ftpFileReader) EOF() (err error) {
	fr.m.Lock()
	if fr.err == nil {
		close(fr.closed)
		fr.err = io.EOF
	}
	err = fr.err
	fr.m.Unlock()
	return
}

func (fr *ftpFileReader) SetError(newErr error) (err error) {
	fr.m.Lock()
	if fr.err == nil {
		close(fr.closed)
		fr.err = io.ErrClosedPipe
	} else {
		err = fr.err
	}
	fr.m.Unlock()
	return
}

func (fr *ftpFileReader) Close() (err error) {
	fr.m.Lock()
	if fr.err == nil {
		close(fr.closed)
		fr.err = io.ErrClosedPipe
	} else {
		err = fr.err
	}
	fr.m.Unlock()
	return nil
}

// Read returns an opened File
func (fs *FtpFilesystem) Read(filepath string) (io.ReadCloser, error) {
	filepath = path.Join(fs.basedir, filepath)

	fmt.Println("Flavor is", fs.flavor)
	if fs.flavor != FTPFlavorVSFTP {
		info, err := fs.client.Stat(filepath)
		if err != nil {
			return nil, ftpConvertError(err)
		}
		if info.IsDir() {
			return nil, fmt.Errorf("Expected a file, found a dir: %s", filepath)
		}
	}

	errChan := make(chan error, 1)

	reader := newFtpFileReader()

	go func() {
		err := fs.client.Retrieve(filepath, reader)
		if err != nil {
			errChan <- err
			reader.SetError(err)
		} else {
			reader.EOF()
		}
	}()

	select {
	case err := <-errChan:
		if fs.flavor == FTPFlavorVSFTP {
			// XXX This is not very reliable, and inspecting the goftp error
			// would bring little added reliability. A more vsftp-adapted
			// client may be the way to go.
			if err.Error() == "unexpected response: 550-Failed to open file." {
				err = os.ErrNotExist
			}
		}
		return nil, err
	case <-reader.Started:
		return reader, nil
	}
}

// Rename renames a file
func (fs *FtpFilesystem) Rename(oldpath, newpath string) error {
	oldpath = path.Join(fs.basedir, oldpath)
	newpath = path.Join(fs.basedir, newpath)
	return ftpConvertError(fs.client.Rename(oldpath, newpath))
}

// Delete removes a file
func (fs *FtpFilesystem) Delete(filepath string) error {
	filepath = path.Join(fs.basedir, filepath)
	return ftpConvertError(fs.client.Delete(filepath))
}

// Observe naïvely do a file list every pollingInterval
func (fs *FtpFilesystem) Observe(
	ctx context.Context, pattern string, logger xbus.Logger,
) (<-chan string, error) {
	patterndir := path.Dir(path.Join(fs.basedir, pattern))

	c := make(chan string)
	fo := NewFileObserver(ctx, pattern, fs.basedir, c)

	doList := func() ([]string, error) {
		infos, err := fs.client.ReadDir(patterndir)
		if ftpErrIsNotExist(err) {
			return nil, nil
		}
		if err != nil {
			log.Printf("ftp: Error listing files: %s", err)
			return nil, err
		}
		var paths []string
		for _, info := range infos {
			if !info.IsDir() {
				paths = append(paths, path.Join(patterndir, info.Name()))
			}
		}
		return paths, nil
	}

	paths, err := doList()
	if err != nil {
		return nil, err
	}

	go func(paths []string) {
		if err := fo.AddFiles(paths...); err != nil {
			log.Printf("ftp: Error emitting files: %s", err)
		}
		var errcount int
		for {
			var sleep time.Duration
			switch errcount {
			case 0:
				sleep = fs.PollingInterval
			case 1:
				sleep = 125 * time.Millisecond
			default:
				sleep *= 2
			}
			if sleep > fs.PollingInterval {
				sleep = fs.PollingInterval
			}
			if errcount == 1 {
				log.Printf("ftp: observe failed, will retry in %s", sleep)
			}
			if errcount > 1 {
				log.Printf("ftp: observe failed %d time, will retry in %s", errcount, sleep)
			}
			select {
			case <-time.After(sleep):
			case <-ctx.Done():
				return
			}
			if paths, err := doList(); err != nil {
				errcount++
			} else {
				if err := fo.AddFiles(paths...); err != nil {
					log.Printf("ftp: Error emitting files: %s", err)
					errcount++
				} else {
					errcount = 0
				}
			}
		}
	}(paths)

	return c, nil
}

// WaitForDeletion wath the given path every second until it does not exist
// anymore
func (fs *FtpFilesystem) WaitForDeletion(ctx context.Context, filepath string) error {
	filepath = path.Join(fs.basedir, filepath)
	for {
		switch fs.flavor {
		case FTPFlavorDefault:
			if _, err := fs.client.Stat(filepath); err != nil {
				if ftpErrIsNotExist(err) {
					return nil
				}
				return err
			}
		case FTPFlavorVSFTP:
			parent, name := path.Split(filepath)
			list, err := fs.client.ReadDir(parent)
			if err != nil {
				return err
			}
			var fileExists bool
			for _, info := range list {
				if info.Name() == name {
					fileExists = true
					break
				}
			}
			if !fileExists {
				return nil
			}
		}
		select {
		case <-time.After(fs.PollingInterval):
		case <-ctx.Done():
			return ctx.Err()
		}
	}
}

func init() {
	MustRegisterFilesystem("ftp", ftpFilesystemFactory)
	MustRegisterFilesystem("ftps", ftpFilesystemFactory)
}
