package filegateway

import (
	"bufio"
	"bytes"
	"context"
	"fmt"
	"io"
	"os"
	"path"
	"strconv"
	"strings"
	"time"

	xbus "xbus.io/go-xbus"
	"golang.org/x/crypto/ssh"
)

func scpFilesystemFactory(options FilesystemOptions) (Filesystem, error) {
	if options.Host == "" {
		return nil, fmt.Errorf(
			"Invalid host for a scp:// uri. Expects a non-mpty host")
	}
	if options.Port == "" {
		options.Port = "22"
	}
	fs := ScpFilesystem{
		FilesystemOptions: options,
		address:           options.Host + ":" + options.Port,
		clientConfig:      ssh.ClientConfig{User: options.Username},
	}
	if options.HasPassword {
		fs.clientConfig.Auth = append(fs.clientConfig.Auth, ssh.Password(options.Password))
	}
	if pkey, err := options.GetCustomStringD("private-key", ""); err == nil {
		if pkey != "" {
			signer, err := ssh.ParsePrivateKey([]byte(pkey))
			if err != nil {
				return nil, fmt.Errorf("Error parsing 'private-key': %s", err)
			}
			fs.clientConfig.Auth = append(fs.clientConfig.Auth, ssh.PublicKeys(signer))
		}
	} else {
		return nil, err
	}
	if hostkey, err := options.GetCustomStringD("host-pubkey", ""); err == nil {
		if hostkey == "" {
			return nil, fmt.Errorf("'host-pubkey' is empty")
		}
		key, _, _, _, err := ssh.ParseAuthorizedKey([]byte(hostkey))
		if err != nil {
			return nil, fmt.Errorf("Error parsing 'host-pubkey': %s", err)
		}
		fs.clientConfig.HostKeyCallback = ssh.FixedHostKey(key)
	} else {
		return nil, err
	}
	if err := options.CustomStringD("basedir", &(fs.basedir), "."); err != nil {
		return nil, err
	}
	return &fs, nil
}

func init() {
	RegisterFilesystem("file", localFilesystemFactory)
}

// ScpFilesystem stores files on the local filesystem
type ScpFilesystem struct {
	FilesystemOptions
	address      string
	clientConfig ssh.ClientConfig
	basedir      string
}

func (fs *ScpFilesystem) getSession() (*ssh.Session, error) {
	client, err := ssh.Dial("tcp", fs.address, &fs.clientConfig)
	if err != nil {
		return nil, fmt.Errorf("Failed to dial: %s", err)
	}
	session, err := client.NewSession()
	if err != nil {
		return nil, fmt.Errorf("Failed to create session: %s", err)
	}
	return session, nil
}

type scpFileWriter struct {
	filename  string
	depth     int
	session   *ssh.Session
	stdin     io.WriteCloser
	runResult chan error
	buffer    *bytes.Buffer
	outBuffer *bytes.Buffer
}

// Write data to scp
func (fw *scpFileWriter) Write(data []byte) (int, error) {
	if fw.buffer != nil {
		return fw.buffer.Write(data)
	}
	return fw.stdin.Write(data)
}

// Close the ssh connection
func (fw *scpFileWriter) Close() error {
	if fw.buffer != nil {
		size := fw.buffer.Len()
		if _, err := fmt.Fprintln(fw.stdin, "C0644", size, fw.filename); err != nil {
			// TODO
			panic(err)
		}
		if _, err := io.Copy(fw.stdin, fw.buffer); err != nil {
			panic(err)
		}
	}
	if _, err := fmt.Fprint(fw.stdin, "\x00"); err != nil {
		panic(err)
	}
	fw.stdin.Close()
	select {
	case err := <-fw.runResult:
		fw.session.Close()
		return err
	case <-time.After(10 * time.Second):
		fw.session.Close()
		return <-fw.runResult
	}
}

// Write ...
func (fs *ScpFilesystem) Write(filepath string, size int) (io.WriteCloser, error) {
	session, err := fs.getSession()
	if err != nil {
		return nil, err
	}
	stdin, err := session.StdinPipe()
	if err != nil {
		session.Close()
		return nil, err
	}
	fw := scpFileWriter{
		filename:  path.Base(filepath),
		session:   session,
		stdin:     stdin,
		runResult: make(chan error),
		outBuffer: bytes.NewBuffer(nil),
	}

	stdout, err := session.StdoutPipe()
	go func() {
		scanner := bufio.NewScanner(stdout)
		for scanner.Scan() {
			fw.outBuffer.Write(scanner.Bytes())
		}
	}()

	stderr, err := session.StderrPipe()
	go func() {
		scanner := bufio.NewScanner(stderr)
		for scanner.Scan() {
			fw.outBuffer.Write(scanner.Bytes())
		}
	}()

	go func() {
		err := session.Run("/usr/bin/scp -tr " + fs.basedir)
		if err != nil {
			err = fmt.Errorf("remote scp failed: %s\nLogs:\n%s", err, fw.outBuffer.String())
		}
		fw.runResult <- err
	}()

	dir := path.Dir(filepath)
	dirpath := strings.Split(dir, "/")
	fw.depth = len(dirpath)
	for _, dir := range dirpath {
		if _, err := fmt.Fprintln(fw.stdin, "D0755", 0, dir); err != nil {
			panic(err)
		}
	}
	if size == -1 {
		fw.buffer = bytes.NewBuffer(nil)
	} else {
		if _, err := fmt.Fprintln(fw.stdin, "C0644", size, fw.filename); err != nil {
			panic(err)
		}
	}
	return &fw, nil
}

type scpFileReader struct {
	header       scpHeader
	remains      int
	session      *ssh.Session
	stdin        io.WriteCloser
	stdoutReader *bufio.Reader
	errBuffer    *bytes.Buffer
	runResult    chan error
}

func (fr *scpFileReader) Read(p []byte) (int, error) {
	if fr.remains == 0 {
		return 0, io.EOF
	}
	if len(p) > fr.remains {
		p = p[:fr.remains]
	}
	n, err := fr.stdoutReader.Read(p)
	if err != nil {
		return n, err
	}
	fr.remains -= n
	return n, nil
}

func (fr *scpFileReader) Close() error {
	fr.stdin.Write([]byte{0})
	fr.stdin.Close()
	select {
	case err := <-fr.runResult:
		fr.session.Close()
		return err
	case <-time.After(10 * time.Second):
		fr.session.Close()
		return <-fr.runResult
	}
}

type scpHeader struct {
	ftype string
	mode  int
	size  int
	fname string
}

func parseScpHeader(line []byte) (*scpHeader, error) {
	if bytes.HasSuffix(line, []byte("\n")) {
		line = line[:len(line)-1]
	}
	splitted := bytes.SplitN(line, []byte(" "), 3)
	if len(splitted) < 3 {
		return nil, fmt.Errorf("Expected 3 fields, got: %s", string(line))
	}
	mode, err := strconv.ParseInt(string(splitted[0][1:]), 8, 16)
	if err != nil {
		return nil, fmt.Errorf("Invalid file mode: %s", err)
	}

	size, err := strconv.Atoi(string(splitted[1]))
	if err != nil {
		return nil, fmt.Errorf("Expected a size: %s", err)
	}

	return &scpHeader{
		ftype: string(splitted[0][:1]),
		mode:  int(mode),
		size:  size,
		fname: string(splitted[2]),
	}, nil
}

// Read returns an opened File
func (fs *ScpFilesystem) Read(filepath string) (io.ReadCloser, error) {
	session, err := fs.getSession()
	if err != nil {
		return nil, err
	}
	stdin, err := session.StdinPipe()
	if err != nil {
		session.Close()
		return nil, err
	}

	stdout, err := session.StdoutPipe()
	if err != nil {
		session.Close()
		return nil, err
	}

	fr := scpFileReader{
		session:      session,
		stdin:        stdin,
		stdoutReader: bufio.NewReader(stdout),
		errBuffer:    bytes.NewBuffer(nil),
		runResult:    make(chan error),
	}

	stderr, err := session.StderrPipe()
	if err != nil {
		session.Close()
		return nil, err
	}

	go func() {
		scanner := bufio.NewScanner(stderr)
		for scanner.Scan() {
			fr.errBuffer.Write(scanner.Bytes())
		}
	}()

	if err := session.Start("/usr/bin/scp -f " + path.Join(fs.basedir, filepath)); err != nil {
		session.Close()
		return nil, err
	}
	go func() {
		err := session.Wait()
		if err != nil {
			err = fmt.Errorf("remote scp failed: %s\nLogs:\n%s", err, fr.errBuffer.String())
		}
		fr.runResult <- err
	}()

	if _, err := stdin.Write([]byte{0}); err != nil {
		session.Close()
		<-fr.runResult
		return nil, err
	}

	line, err := fr.stdoutReader.ReadBytes('\n')
	if err != nil {
		session.Close()
		<-fr.runResult
		return nil, err
	}
	if bytes.Contains(line, []byte("No such file or directory")) {
		session.Close()
		<-fr.runResult
		return nil, os.ErrNotExist
	}
	header, err := parseScpHeader(line)
	if err != nil {
		session.Close()
		<-fr.runResult
		return nil, err
	}
	if _, err := stdin.Write([]byte{0}); err != nil {
		session.Close()
		<-fr.runResult
		return nil, err
	}
	fr.header = *header
	fr.remains = header.size

	return &fr, nil
}

// Rename renames a file
func (fs *ScpFilesystem) Rename(oldpath, newpath string) error {
	session, err := fs.getSession()
	if err != nil {
		return err
	}
	defer session.Close()
	if output, err := session.CombinedOutput("mv " +
		path.Join(fs.basedir, oldpath) + " " +
		path.Join(fs.basedir, newpath)); err != nil {
		if bytes.Contains(output, []byte("No such file or directory")) {
			return os.ErrNotExist
		}
		return fmt.Errorf("mv failed: %s\n%s", err, string(output))
	}
	return nil
}

// Delete removes a file
func (fs *ScpFilesystem) Delete(filepath string) error {
	session, err := fs.getSession()
	if err != nil {
		return err
	}
	defer session.Close()
	if output, err := session.CombinedOutput("rm " + path.Join(fs.basedir, filepath)); err != nil {
		if bytes.Contains(output, []byte("No such file or directory")) {
			return os.ErrNotExist
		}
		return fmt.Errorf("rm failed: %s\n%s", err, string(output))
	}
	return nil
}

// Observe naïvely do a file list every pollingInterval
func (fs *ScpFilesystem) Observe(
	ctx context.Context, pattern string, logger xbus.Logger,
) (<-chan string, error) {
	c := make(chan string)
	fo := NewFileObserver(ctx, pattern, fs.basedir, c)

	session, err := fs.getSession()
	if err != nil {
		return nil, err
	}
	session.RequestPty("", 25, 80, nil)
	stdout, err := session.StdoutPipe()
	if err != nil {
		session.Close()
		return nil, err
	}

	stderr, err := session.StderrPipe()
	if err != nil {
		session.Close()
		return nil, err
	}
	errBuffer := bytes.NewBuffer(nil)
	go io.Copy(errBuffer, stderr)

	cmd := fmt.Sprintf(
		"while (ls --color=never %s || true); do sleep %.4f; done",
		path.Join(fs.basedir, pattern),
		fs.PollingInterval.Seconds())
	if err := session.Start(cmd); err != nil {
		session.Close()
		return nil, err
	}

	result := make(chan error)
	go func() {
		err := session.Wait()
		if err != nil {
			err = fmt.Errorf("observe failed: %s\nLogs:\n%s", err, errBuffer.String())
		} else {
			err = fmt.Errorf("observe stopped: Logs:\n%s", errBuffer.String())
		}
		result <- err
	}()

	go func() {
		scanner := bufio.NewScanner(stdout)
		for scanner.Scan() {
			fmt.Println("observe:", scanner.Text())
			fo.AddFiles(scanner.Text())
		}
	}()

	go func() {
		select {
		case err := <-result:
			session.Close()
			logger.Errorf("scp: %s", err)
		case <-ctx.Done():
			session.Close()
			<-result
		}
	}()
	return c, nil
}

// WaitForDeletion wath the given path every second until it does not exist
// anymore
func (fs *ScpFilesystem) WaitForDeletion(ctx context.Context, filepath string) error {
	session, err := fs.getSession()
	if err != nil {
		return err
	}

	result := make(chan error)

	go func() {
		if output, err := session.CombinedOutput(
			fmt.Sprintf("while ( ls '%s' ); do sleep %.4f; done",
				path.Join(fs.basedir, filepath),
				fs.PollingInterval.Seconds()),
		); err != nil {
			result <- fmt.Errorf("wait deletion failed: %s\n%s", err, string(output))
		} else {
			result <- nil
		}
	}()

	select {
	case err := <-result:
		session.Close()
		return err
	case <-ctx.Done():
		session.Close()
		<-result
		return ctx.Err()
	}
}

func init() {
	MustRegisterFilesystem("scp", scpFilesystemFactory)
}
